package Mentat::Reporter::Fetcher::Default;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Reporter::Fetcher::Default

=head1 SYNOPSIS

 use Mentat::Reporter::Fetcher::Default;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Reporter::Fetcher(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Time::HiRes qw(time);

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Mentat::Reporter::Fetcher;
use Value::Convertor;
use Data::Filter::Factory;
use Data::Filter;
use Mentat::Message::Value::IDEA;
use Mentat::Storage::Mongo;
use Mentat::Message::Convertor::Mongo;
use Mentat::Message::Storage::Mongo;
use Mentat::Message::Query::Mongo;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Reporter::Fetcher');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item fetch_abuse_list() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub fetch_abuse_list()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $query_string = $self->_generate_query_string();
    print STDERR "FAL query: '$query_string'\n" if $DEVEL;
    my $query_hash = $self->{QUERY}->generate_query($self->{FFACTORY}->filter_from_string($query_string));

    return $self->{STORAGE}->distinct('_CESNET.ResolvedAbuses', 'alerts', $query_hash);
}

=item get_severity_list() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub get_severity_list()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Lazy data preloading
    $self->_preload_data();

    return [keys(%{$self->{LIST_SEVERITIES}})];
}

=item get_abuse_list() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub get_abuse_list()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Lazy data preloading
    $self->_preload_data();

    return [keys(%{$self->{LIST_ABUSES}})];
}

=item get_data($$$$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub get_data($$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($tl, $th, $abuse, $severity) = @_;

    # Lazy data preloading
    $self->_preload_data();

    return $self->_filter_data($self->{DATA}, $tl, $th, $abuse, $severity);
}

=item get_abuse_data($$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub get_abuse_data($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($tl, $th, $abuse) = @_;

    # Lazy data preloading
    $self->_preload_data();

    return $self->_filter_abuse_data($self->{DATA}, $tl, $th, $abuse);
}

=item get_raw_data($$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub get_raw_data($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($tl, $th) = @_;

    # Lazy data preloading
    $self->_preload_data();

    return $self->_filter_raw_data($self->{DATA_RAW}, $tl, $th);
}

=item get_all_data() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub get_all_data()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Lazy data preloading
    $self->_preload_data();

    return $self->{DATA};
}

=item get_status() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub get_status(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return $self->{STATUS};
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Reporter::Fetcher::Default instance
#  Arguments : Unknown see the description in appropriate subclass
#  Returns   : Mentat::Reporter::Fetcher::Default $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat::Reporter::Fetcher::Default instance...
    my %config = @_;

    die "[ENG:Reporter:Fetcher] Filter grammar file must be specified under configuration key 'filter_grammar_file'\n" unless exists $config{filter_grammar_file};
    die "[ENG:Reporter:Fetcher] Filter grammar file '$config{filter_grammar_file}' must be ordinary file\n" unless -f $config{filter_grammar_file};
    die "[ENG:Reporter:Fetcher] Filter schema file must be specified under configuration key 'filter_schema_file'\n" unless exists $config{filter_schema_file};
    die "[ENG:Reporter:Fetcher] Filter schema file '$config{filter_schema_file}' must be ordinary file\n" unless -f $config{filter_schema_file};
    die "[ENG:Reporter:Fetcher] Prefilter engine configuration 'filter_config' is missing\n" unless $config{filter_config};

    $self->{FFACTORY}  = Data::Filter::Factory->new_from_file($config{filter_grammar_file});
    $self->{CONVERTOR} = Mentat::Message::Convertor::Mongo->new();
    $self->{STORAGE}   = Mentat::Message::Storage::Mongo->get();
    $self->{QUERY}     = Mentat::Message::Query::Mongo->new(convertor => $self->{CONVERTOR}, storage => $self->{STORAGE});

    $self->{TL}    = $config{tl};
    $self->{TH}    = $config{th};
    $self->{ABUSE} = $config{abuse};

    $self->{PREFILTER} = Data::Filter->new(filter_grammar_file => $config{filter_grammar_file},
                                           filter_schema_file  => $config{filter_schema_file},
                                           filter_config       => $config{filter_config},
                                           dt_resolver         => Mentat::Message::Value::IDEA->new());

    my $eventdb = Mentat::Storage::Mongo->new(database => 'mentat', collection => 'events');
    my ($events, $event_cnt) = $eventdb->find();

    if ($event_cnt) {
        foreach my $e (@$events) {
            $self->{EVENTS}->{$e->{_id}} = $e->{severity};
        }
    }

    return $self;
}

# _generate_query_string() [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _generate_query_string()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $query_string = '(ts ge "'.$self->{TL}.'") and (ts le "'.$self->{TH}.'")';
    unless ($self->{ABUSE}) {
        $query_string .= ' and (EXISTS _CESNET/ResolvedAbuses)';
    }
    else {
        $query_string .= ' and (_CESNET/ResolvedAbuses eq "'.$self->{ABUSE}.'")';
    }
    return $query_string;
}

# _prefilter_messages($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _prefilter_messages($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($messages,) = @_;
    Log::Loger->debug(__PACKAGE__, "[ENG:Reporter:Fetcher] Prefiltering ".(scalar @$messages)." messages");

    my $tmp   = [];
    my $cntb  = 0;
    my $cntp  = 0;
    my $start = time;
    my ($result, $rule, $idx, $m, $al);
    foreach my $msg (@$messages) {
        eval {
            $m = $msg->[0];
            $al = $m->path_values("_CESNET/ResolvedAbuses");
            ($result, $rule, $idx) = $self->{PREFILTER}->filter($m);

            # Message denied
            if ($result == 0) {
                #Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Fetcher] Message '".$m->id()."' was blocked by filter according to rule '$rule:$idx'");

                $cntb++;
                foreach my $a (@$al) {
                    $self->{STATUS}->{structure}->{$a}->{prefilter}->{$rule}++;
                }
                $self->{STATUS}->{prefilter}->{$rule}++;
            }
            # Message allowed
            elsif ($result == 1) {
                push(@$tmp, $msg);
                $cntp++;
                foreach my $a (@$al) {
                    $self->{STATUS}->{structure}->{$a}->{postfilter}++;
                }
            }
        };
        if ($@) {
            $self->{STATUS}->{errors}->{prefilter}++;
            Log::Loger->error(__PACKAGE__, "[ENG:Reporter:Fetcher] Prefiltering error (skipping message): $@");
        }
    }
    $messages = $tmp;
    my $stop = time;

    $self->{STATUS}->{counts}->{flt} = $cntp;
    $self->{STATUS}->{counts}->{blk} = $cntb;

    $self->{STATUS}->{durations}->{'prefilter'} = $stop - $start;

    Log::Loger->notice(__PACKAGE__, "[ENG:Reporter:Fetcher] Prefiltering complete, '$cntb' messages blocked, '$cntp' messages passed (t: ".($stop - $start)." s)");
    return $messages;
}

# _preload_data(;$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _preload_data(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($force,) = @_;

    return if $self->{DATA_RAW} and not $force;

    my $start = time;
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Fetcher] Preloading data for time window '".(scalar localtime($self->{TL}))." - ".(scalar localtime($self->{TH}))."' = '$self->{TL}:$self->{TH}' (".(Value::Convertor->duration_to_string($self->{TH} - $self->{TL})).")");
    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Fetcher] Preloading data for specific abuse group '$self->{ABUSE}'") if $self->{ABUSE};

    my $query_string = $self->_generate_query_string();
    print STDERR "PLD query: '$query_string'\n" if $DEVEL;
    my $query_hash = $self->{QUERY}->generate_query($self->{FFACTORY}->filter_from_string($query_string));

    my ($messages, $count_all);
    eval {
        ($messages, $count_all) = $self->{QUERY}->query_msg_json_wm($query_hash, {'ts' => -1}, 1000000);
    };
    if ($@) {
        Log::Loger->error(__PACKAGE__, "[ENG:Reporter:Fetcher] $@");
        die $@;
    }

    $self->{DATA_RAW} = $messages;
    $self->{STATUS}->{counts}    = {'all'   => $count_all};
    $self->{STATUS}->{durations} = {'fetch' => (time - $start)};

    if ($self->{STATUS}->{counts}->{'all'}) {
        $self->{DATA_RAW} = $self->_prefilter_messages($self->{DATA_RAW});
    }

    if ($self->{STATUS}->{counts}->{'flt'}) {
        #my $previous;
        my ($abuses, $templates, $severity);
        foreach my $msg (@{$self->{DATA_RAW}}) {
            #print STDERR "ID: $msg->[1]->{_id}, TSU: $msg->[1]->{_CESNET}->{StorageTime}\n";
            #die "Unsorted result set received '$previous < $msg->[1]->{_CESNET}->{StorageTime}'\n" if $previous and $previous < $msg->[1]->{_CESNET}->{StorageTime};
            #$previous = $msg->[1]->{_CESNET}->{StorageTime};

            $abuses    = $msg->[0]->path_values("_CESNET/ResolvedAbuses");
            $templates = $msg->[0]->path_values("_CESNET/EventTemplate");
            $templates = ['__unknown__'] unless @$templates;

            foreach my $a (@$abuses) {
                foreach my $t (@$templates) {
                    $severity = $self->{EVENTS}->{$t};
                    $severity = 'low' unless $severity;

                    push(@{$self->{DATA}->{$a}->{$severity}}, $msg);
                    push(@{$self->{DATA}->{$a}->{all}}, $msg);

                    $self->{LIST_SEVERITIES}->{$severity}++;
                    $self->{LIST_ABUSES}->{$a}++;

                    $self->{STATUS}->{abuses}->{$a}++;
                    $self->{STATUS}->{templates}->{$t}++;
                    $self->{STATUS}->{severities}->{$severity}++;

                    $self->{STATUS}->{structure}->{$a}->{templates}->{$t}++;
                    $self->{STATUS}->{structure}->{$a}->{severities}->{$severity}++;
                }
            }
        }
    }

    my $stop = time;
    $self->{STATUS}->{durations}->{'total'} = $stop - $start;

    Log::Loger->info(__PACKAGE__, "[ENG:Reporter:Fetcher] Preloaded data for time window '".(scalar localtime($self->{TL}))." - ".(scalar localtime($self->{TH}))."' = '$self->{TL}:$self->{TH}' [".($stop - $start)." s, ".($self->{STATUS}->{counts}->{'flt'}||0)." #]");
    #Log::Loger->debug(__PACKAGE__, "PRELOADED DATA: " . Dumper($self->{DATA})) if $DEVEL;
}

# _filter_data($$$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _filter_data($$$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($data, $tl, $th, $abuse, $severity) = @_;

    my $result = [];
    return $result unless exists $data->{$abuse} and exists $data->{$abuse}->{$severity};

    foreach my $d (@{$data->{$abuse}->{$severity}}) {
        next if $d->[1]->{_CESNET}->{StorageTime} > $th;
        last if $d->[1]->{_CESNET}->{StorageTime} < $tl;

        push(@{$result}, $d->[0]);
    }
    return $result;
}

# _filter_abuse_data($$$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _filter_abuse_data($$$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($data, $tl, $th, $abuse) = @_;

    my $result = [];

    foreach my $d (@{$data->{$abuse}->{all}}) {
        next if $d->[1]->{_CESNET}->{StorageTime} > $th;
        last if $d->[1]->{_CESNET}->{StorageTime} < $tl;

        push(@{$result}, $d->[0]);
    }
    return $result;
}

# _filter_raw_data($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _filter_raw_data($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($data, $tl, $th) = @_;

    my $result = [];

    foreach my $d (@$data) {
        next if $d->[1]->{_CESNET}->{StorageTime} > $th;
        last if $d->[1]->{_CESNET}->{StorageTime} < $tl;

        push(@{$result}, $d->[0]);
    }
    return $result;
}

=pod

=back

=cut

1;
