package Whois::Dump;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Whois::Dump - Simple whois database constructed by parsing whois dump file

=head1 SYNOPSIS

 use Whois::Dump;

 # Instantinate the whois database by parsing given whois dump, file must not be
 # older than three days
 my $wi_tree = Whois::Dump->new('./data/whois.dump');

 # Instantinate the whois database by parsing given whois dump, do not check the
 # file age
 my $wi_tree = Whois::Dump->new('./data/whois.dump', 1);

 # Find information regarding the given IP address
 my @results = $wi_tree->find('146.102.1.1');

 # Find information regarding the given IP address
 my $abuse = $wi_tree->find_abuse('146.102.1.1');

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Value::Convertor;
use Value::IPv4ADDR;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant MAX_AGE => 259200;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.6';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC ,STATIC]

  Usage     : my $instance = Whois::Dump->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Whois::Dump reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item check() [PUBLIC]

  Usage     : my @warnings = $instance->check()
  Purpose   : Check the inner whois database for errors and warnings
  Arguments : None
  Returns   : ARRAY containing warnings
  Throws    : Dies, if not invoked on object instance

=cut

sub check()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my @warnings = ();
    foreach my $key (keys(%{$self->{RANGES}})) {
        # 'abuse-mailbox' should be always present
        unless (exists $self->{RANGES}->{$key}->{'abuse-mailbox'}) {
            # Suppress this warning
            #push(@warnings, "Network '$key' does not have 'abuse-mailbox' contact defined'");

            # If 'abuse-mailbox' does not exist, look into remarks
            unless (exists $self->{RANGES}->{$key}->{'remarks'}) {
                push(@warnings, "Network '$key' does not have abuse contact defined in either 'abuse-mailbox' or 'remarks' attributes");

                # Last option are contact emails for technical, or admin contact persons
                unless (exists $self->{RANGES}->{$key}->{'e-mail'}) {
                    push(@warnings, "Network '$key' does not have any abuse contact defined'");
                }
            }
        }
    }
    return @warnings;
}

=item find($) [PUBLIC]

  Usage     : my $result = $instance->find('146.102.1.1');
  Purpose   : Search for given IPv4 address in the inner whois storage
  Arguments : MIXED $ipv4 - IPv4 address
  Returns   : ARRAY OF HASH REFERENCES containing range structures
  Throws    : Dies, if not invoked on object instance

=cut

sub find($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ipv4,) = @_;

    my $ipv4obj = Value::IPv4ADDR->new($ipv4);
    my $ipv4num = $ipv4obj->to_numeric();

    my @ranges = ();
    foreach my $range (keys(%{$self->{RANGES}})) {
        push(@ranges, $self->{RANGES}->{$range}) if $self->_ip_in_range($ipv4num, $self->{RANGES}->{$range});
        $self->{RANGES}->{$range}->{source} = 'whois-dump';
    }
    return  sort { $b->{'mask'} <=> $a->{'mask'} }
            sort { $b->{'pref'} <=> $a->{'pref'} } @ranges;
}

=item find_abuse($) [PUBLIC]

  Usage     : my $result = $instance->find_abuse('146.102.1.1');
  Purpose   : Search for abuse contact for given IPv4 address in the inner whois storage
  Arguments : Same as find() method
  Returns   : STRING $abuse contact on SUCCESS, UNDEF on FAILURE
  Throws    : Dies, if not invoked on object instance
  See also  : find() method

=cut

sub find_abuse($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Check the internal whois storage
    my @results = $self->find(@_);
    if (@results) {
        my (@abuse_mailboxes, @remarks, @emails);
        @abuse_mailboxes = @{$results[0]->{'abuse-mailbox'}} if ref $results[0]->{'abuse-mailbox'};
        @remarks         = @{$results[0]->{'remarks'}} if ref $results[0]->{'remarks'};
        @emails          = @{$results[0]->{'e-mail'}} if ref $results[0]->{'e-mail'};

        if (@abuse_mailboxes) {
            return join(',', @abuse_mailboxes);
        }
        elsif (@remarks) {
            return join(',', @remarks);
        }
        #elsif (@emails) {
        #   return join(',', @emails);
        #}
        else {
            return '';
        }
    }
    return undef;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init($) [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Whois::Dump instance
#  Arguments : STRING $whois_file - name of the file containing whois dump
#              BOOL $force - parse the whois dump file, even if it is too old
#  Returns   : Whois::Dump reference
#  Throws    : Dies, if not invoked on object instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($whois_file, $force) = @_;
    die "Static whois file must be given as argument" unless $whois_file;
    die "Static whois file '$whois_file' must be ordinary file" unless -f $whois_file;

    # We will be using convertor for IP address conversions
    $self->{CONVERTOR} = Value::Convertor->new();

    # Storage for IP ranges
    $self->{RANGES} = {};

    # Check the age of the whois dump file
    my $ts = time();
    my $ts_w = (stat($whois_file))[9];
    if (($ts - $ts_w) > MAX_AGE()) {
        unless ($force) {
            die "Static whois dump file '$whois_file' is older than 3 days\n";
        }
        elsif ($force == 1) {
            warn "Static whois dump file '$whois_file' is older than 3 days\n";
        }
    }

    # Load the whois database from given whois dump file
    $self->_read_file($whois_file);

    # Prepare the inner database to better format
    $self->_prepare();

    return $self;
}

# _prepare() [PROTECTED]
#
#  Usage     : $self->_prepare();
#  Purpose   : Prepare the inner database after loading from external file
#  Arguments : None
#  Returns   : Nothing
#  Throws    : Dies, if not invoked on object instance

sub _prepare()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    foreach my $key (keys(%{$self->{RANGES}})) {
        if (exists $self->{RANGES}->{$key}->{'abuse-mailbox'}) {
            $self->{RANGES}->{$key}->{'abuse-mailbox'} = [keys(%{$self->{RANGES}->{$key}->{'abuse-mailbox'}})];
        }
        if (exists $self->{RANGES}->{$key}->{'remarks'}) {
            $self->{RANGES}->{$key}->{'remarks'} = [keys(%{$self->{RANGES}->{$key}->{'remarks'}})];
        }
        if (exists $self->{RANGES}->{$key}->{'e-mail'}) {
            $self->{RANGES}->{$key}->{'e-mail'} = [keys(%{$self->{RANGES}->{$key}->{'e-mail'}})];
        }
    }
}

# _read_file($) [PROTECTED]
#
#  Usage     : $self->_read_file($whois_file);
#  Purpose   : Populate the inner database by parsing the external whois dump file
#  Arguments : STRING $whois_file - name of the file containing whois dump
#  Returns   : Nothing
#  Throws    : Dies, if not invoked on object instance

sub _read_file($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($whois_file,) = @_;

    # Context for passing between the calls for storing results
    my $context;

    open(my $input, $whois_file) or die "Can`t open whois file '$whois_file'\n";
    while (my $line = <$input>) {
        # Remove the whitespaces
        chomp($line);

        # Analyze the line
        $context = $self->_analyze_line($context, $line);
    }
    close($input);
}

# _analyze_line($$) [PROTECTED]
#
#  Usage     : $context = $self->_analyze_line($context, $line);
#  Purpose   : Analyze given line within given context
#  Arguments : HASH REFERENCE $context - current context
#              STRING $line - one line from the whois dump file
#  Returns   : HASH REFERENCE $context
#  Throws    : Dies, if not invoked on object instance

sub _analyze_line($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($context, $line) = @_;

    # Skip empty lines
    return $context unless $line;

    # Section switching - range delimiter
    if ($line =~ /^% Information related to '([\d]+\.[\d]+\.[\d]+\.[\d]+) - ([\d]+\.[\d]+\.[\d]+\.[\d]+)'/) {
        my $v1i = $self->{CONVERTOR}->ipv4str_to_int($1);
        my $v2i = $self->{CONVERTOR}->ipv4str_to_int($2);
        return "$1:$2|$v1i:$v2i";
    }
    # Section switching - range delimiter
    elsif ($line =~ /^% Information related to '([\d]+\.[\d]+\.[\d]+\.[\d]+)\/([\d]+)AS2852/) {
        my $v1i = $self->{CONVERTOR}->ipv4str_to_int($1);
        my $v2i = ($v1i + (2 ** (32 - $2))) - 1;
        return "$1/$2|$v1i:$v2i";
    }
    # Skip other comment lines
    elsif ($line =~ /^%/) {
        return $context;
    }
    # Following checks do not make sense on empty context
    elsif ($context) {
        # Perform some argument preparation first
        my ($key, $value) = split(/:/, $line);
        if ($key) {
            $key =~ s/^\s+//; $key =~ s/\s+$//;
        }
        if ($value) {
            $value =~ s/^\s+//; $value =~ s/\s+$//;
        }
        $value = $key   if not $value and $key;

        # 'netname' does not need any preprocessing
        if ($key =~ /^netname$/) {
            $self->{RANGES}->{$context}->{$key} = $value;
        }
        # 'route' attribute - compute range intervals
        elsif ($key =~ /^route$/) {
            my ($ip_min_str, $mask) = split(/[\/]/, $value);

            my $ip_min_int  = $self->{CONVERTOR}->ipv4str_to_int($ip_min_str);
            my $size        = 2 ** (32 - $mask);
            my $ip_max_int  = ($ip_min_int + $size) - 1;

            $self->{RANGES}->{$context}->{'route'}          = $value;
            $self->{RANGES}->{$context}->{'route_int'}      = "$ip_min_int:$ip_max_int";
            $self->{RANGES}->{$context}->{'ip_min_int'}     = $ip_min_int;
            $self->{RANGES}->{$context}->{'ip_max_int'}     = $ip_max_int;
            $self->{RANGES}->{$context}->{'ip_min_str'}     = $self->{CONVERTOR}->int_to_ipv4str($ip_min_int);
            $self->{RANGES}->{$context}->{'ip_max_str'}     = $self->{CONVERTOR}->int_to_ipv4str($ip_max_int);
            $self->{RANGES}->{$context}->{'size'}           = $size;
            $self->{RANGES}->{$context}->{'mask'}           = $mask;
            $self->{RANGES}->{$context}->{'pref'}           = 0;
        }
        # 'inetnum' attribute - compute range intervals
        elsif ($key =~ /^inetnum$/) {
            $value =~ s/[- \t]+/:/;
            my ($ip_min_str, $ip_max_str) = split(/[:]/, $value);

            my $ip_min_int = $self->{CONVERTOR}->ipv4str_to_int($ip_min_str);
            my $ip_max_int = $self->{CONVERTOR}->ipv4str_to_int($ip_max_str);

            $self->{RANGES}->{$context}->{'inetnum'}        = $value;
            $self->{RANGES}->{$context}->{'inetnum_int'}    = "$ip_min_int:$ip_max_int";
            $self->{RANGES}->{$context}->{'ip_min_int'}     = $ip_min_int;
            $self->{RANGES}->{$context}->{'ip_max_int'}     = $ip_max_int;
            $self->{RANGES}->{$context}->{'ip_min_str'}     = $ip_min_str;
            $self->{RANGES}->{$context}->{'ip_max_str'}     = $ip_max_str;
            $self->{RANGES}->{$context}->{'size'}           = ($ip_max_int - $ip_min_int) + 1;
            $self->{RANGES}->{$context}->{'mask'}           = 32 - log(($ip_max_int - $ip_min_int) + 1)/log(2);
            $self->{RANGES}->{$context}->{'pref'}           = 1;
        }
        # Stack email addresses (perform 'uniq')
        elsif ($key =~ /^e-mail|abuse-mailbox$/) {
            $self->{RANGES}->{$context}->{$key}->{$value}++;
        }
        # Parse out and stack all email addresses from 'remarks' attribute
        elsif (($key =~ /^remarks$/) and ($value =~ /abuse/)) {
            my @matches = $value =~ /([^\s]+@\w+(?:\.\w+)+)/g;
            map { $self->{RANGES}->{$context}->{$key}->{$_}++; } @matches;
        }
    }
    return $context;
}

# _ip_in_range($$) [PROTECTED]
#
#  Usage     : my $result = $self->_ip_in_range($ipv4, $self->{RANGES}->{$range});
#  Purpose   : Check, if the given IPv4 address is within the given IPv4 range
#  Arguments : INTEGER $ipv4int - IPv4 address as 32bit integer
#              HASH REFERENCE $range - range structure from inner storage
#  Returns   : INTEGER 1 on success, INTEGER 0 on failure
#  Throws    : Dies, if not invoked on object instance

sub _ip_in_range($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ipv4int, $range) = @_;

    return 0 if $ipv4int < $range->{'ip_min_int'};
    return 0 if $ipv4int > $range->{'ip_max_int'};
    return 1;
}

=pod

=back

=cut

1;
